from datetime import datetime
import os
import sys


extensions = ["releases"]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
exclude_patterns = ["_build"]

project = u"pytest-relaxed"
year = datetime.now().year
copyright = u"%d Jeff Forcier" % year

# Ensure project directory is on PYTHONPATH for version, autodoc access
sys.path.insert(0, os.path.abspath(os.path.join(os.getcwd(), "..")))

# Extension/theme settings
html_theme_options = {
    "description": "Relaxed pytest discovery",
    "github_user": "bitprophet",
    "github_repo": "pytest-relaxed",
    "fixed_sidebar": True,
}
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
        "donate.html",
    ]
}
# TODO: make it easier to configure this and alabaster at same time, jeez
releases_github_path = "bitprophet/pytest-relaxed"



import pytest

from .classes import SpecModule
from .reporter import RelaxedReporter

# NOTE: fixtures must be present in the module listed under our setup.py's
# pytest11 entry_points value (i.e., this one.) Just being in the import path
# (e.g. package __init__.py) was not sufficient!
from .fixtures import environ  # noqa


def pytest_ignore_collect(path, config):
    # Ignore files and/or directories marked as private via Python convention.
    return path.basename.startswith("_")


def pytest_collect_file(path, parent):
    # Modify file selection to choose all .py files besides conftest.py.
    # (Skipping underscored names is handled up in pytest_ignore_collect, which
    # applies to directories too.)
    if (
        path.ext == ".py"
        and path.basename != "conftest.py"
        and
        # Also skip anything prefixed with test_; pytest's own native
        # collection will get that stuff, and we don't _want_ to try modifying
        # such files anyways.
        not path.basename.startswith("test_")
    ):
        # Then use our custom module class which performs modified
        # function/class selection as well as class recursion
        return SpecModule(path, parent)


@pytest.mark.trylast  # So we can be sure builtin terminalreporter exists
def pytest_configure(config):
    # TODO: we _may_ sometime want to do the isatty/slaveinput/etc checks that
    # pytest-sugar does?
    builtin = config.pluginmanager.getplugin("terminalreporter")
    # Pass the configured, instantiated builtin terminal reporter to our
    # instance so it can refer to e.g. the builtin reporter's configuration
    ours = RelaxedReporter(builtin)
    # Unregister the builtin first so only our output appears
    config.pluginmanager.unregister(builtin)
    config.pluginmanager.register(ours, "terminalreporter")



import re

from _pytest.terminal import TerminalReporter


# TODO:
# - how can we be sure the tests are in the right order?
#   - aka how can we 'sort' them? in post-collection step?
# - how to handle display of 'header' lines? Probably state tracking as w/
# spec?
# - how to deal with flat modules vs nested classes?
# - would be nice to examine all tests in a tree, but that requires waiting
# till all results are in, which is no bueno. So we really do just need to
# ensure a tree-based sort (which, assuming solid test ID strings, can be a
# lexical sort.)
#   - sadly, what this means is that the parent/child relationship between test
#   objects doesn't really help us any, since we have to take action on a
#   per-report basis. Meh. (guess if we NEEDED to access attributes of a parent
#   in a child, that'd be possible, but...seems unlikely-ish? Maybe indent
#   based on parent relationships instead of across-the-run state tracking?)


TEST_PREFIX = re.compile(r"^(Test|test_)")
TEST_SUFFIX = re.compile(r"(Test|_test)$")


# NOTE: much of the high level "replace default output bits" approach is
# cribbed directly from pytest-sugar at 0.8.0
class RelaxedReporter(TerminalReporter):

    def __init__(self, builtin):
        # Pass in the builtin reporter's config so we're not redoing all of its
        # initial setup/cli parsing/etc. NOTE: TerminalReporter is old-style :(
        TerminalReporter.__init__(self, builtin.config)
        # Which headers have already been displayed
        # TODO: faster data structure probably wise
        self.headers_displayed = []
        # Size of indents. TODO: configuration
        self.indent = " " * 4

    def pytest_runtest_logstart(self, nodeid, location):
        # Non-verbose: do whatever normal pytest does.
        if not self.verbosity:
            return TerminalReporter.pytest_runtest_logstart(
                self, nodeid, location
            )
        # Verbose: do nothing, preventing normal display of test location/id.
        # Leaves all display up to other hooks.

    def pytest_runtest_logreport(self, report):
        # TODO: if we _need_ access to the test item/node itself, we may want
        # to implement pytest_runtest_makereport instead? (Feels a little
        # 'off', but without other grody hax, no real way to get the obj so...)

        # Non-verbose: do whatever normal pytest does.
        # TODO: kinda want colors & per-module headers/indent though...
        if not self.verbosity:
            return TerminalReporter.pytest_runtest_logreport(self, report)

        # First, the default impl of this method seems to take care of updating
        # overall run stats; if we don't repeat that we lose all end-of-run
        # tallying and whether the run failed...kind of important. (Why that's
        # not a separate hook, no idea :()
        self.update_stats(report)
        # After that, short-circuit if it's not reporting the main call (i.e.
        # we don't want to display "the test" during its setup or teardown)
        if report.when != "call":
            return
        id_ = report.nodeid
        # First, make sure we display non-per-test data, i.e.
        # module/class/nested class headers (which by necessity also includes
        # tracking indentation state.)
        self.ensure_headers(id_)
        # Then we can display the test name/status itself.
        self.display_result(report)

    def update_stats(self, report):
        status_getter = self.config.hook.pytest_report_teststatus
        cat, letter, word = status_getter(report=report)
        self.stats.setdefault(cat, []).append(report)
        # For use later; apparently some other plugins can yield display markup
        # in the 'word' field of a report.
        self.report_word = word

    def split(self, id_):
        # Split on pytest's :: joiner, and strip out our intermediate
        # SpecInstance objects (appear as '()')
        headers = [x for x in id_.split("::")[1:] if x != "()"]
        # Last one is the actual test being reported on, not a header
        leaf = headers.pop()
        return headers, leaf

    def transform_name(self, name):
        """
        Take a test class/module/function name and make it human-presentable.
        """
        # TestPrefixes / test_prefixes -> stripped
        name = re.sub(TEST_PREFIX, "", name)
        # TestSuffixes / suffixed_test -> stripped
        name = re.sub(TEST_SUFFIX, "", name)
        # All underscores become spaces, for sentence-ishness
        name = name.replace("_", " ")
        return name

    def ensure_headers(self, id_):
        headers, _ = self.split(id_)
        printed = False
        # TODO: this works for class-based tests but needs love for module ones
        # TODO: worth displaying filename ever?
        # Make sure we print all not-yet-seen headers
        for i, header in enumerate(headers):
            # Need to semi-uniq headers by their 'path'. (This is a lot like
            # "the test id minus the last segment" but since we have to
            # split/join either way...whatever. I like dots.)
            header_path = ".".join(headers[: i + 1])
            if header_path in self.headers_displayed:
                continue
            self.headers_displayed.append(header_path)
            indent = self.indent * i
            header = self.transform_name(header)
            self._tw.write("\n{}{}\n".format(indent, header))
            printed = True
        # No trailing blank line after all headers; only the 'last' one (i.e.
        # before any actual test names are printed). And only if at least one
        # header was actually printed! (Otherwise one gets newlines between all
        # tests.)
        if printed:
            self._tw.write("\n")

    def display_result(self, report):
        headers, leaf = self.split(report.nodeid)
        indent = self.indent * len(headers)
        leaf = self.transform_name(leaf)
        # This _tw.write() stuff seems to be how vanilla pytest writes its
        # colorized verbose output. Bit clunky, but it means we automatically
        # honor things like `--color=no` and whatnot.
        self._tw.write(indent)
        self._tw.write(leaf, **self.report_markup(report))
        self._tw.write("\n")

    def report_markup(self, report):
        # Basically preserved from parent implementation; if something caused
        # the 'word' field in the report to be a tuple, it's a (word, markup)
        # tuple. We don't care about the word (possibly bad, but it doesn't fit
        # with our display ethos right now) but the markup may be worth
        # preserving.
        if isinstance(self.report_word, tuple):
            return self.report_word[1]
        # Otherwise, assume ye olde pass/fail/skip.
        if report.passed:
            color = "green"
        elif report.failed:
            color = "red"
        elif report.skipped:
            color = "yellow"
        return {color: True}



import pytest

from pytest_relaxed import raises


class Boom(Exception):
    pass


class OtherBoom(Exception):
    pass


class Test_raises(object):

    def test_when_given_exception_raised_no_problem(self):

        @raises(Boom)
        def kaboom():
            raise Boom

        kaboom()
        # If we got here, we're good...

    def test_when_given_exception_not_raised_it_raises_Exception(self):
        # TODO: maybe raise a custom exception later? HEH.
        @raises(Boom)
        def kaboom():
            pass

        # Buffalo buffalo
        with pytest.raises(Exception) as exc:
            kaboom()
        assert "Did not receive expected Boom!" in str(exc.value)

    def test_when_some_other_exception_raised_it_is_untouched(self):

        @raises(Boom)
        def kaboom():
            raise OtherBoom("sup")

        # Buffalo buffalo
        with pytest.raises(OtherBoom) as exc:
            kaboom()
        assert "sup" == str(exc.value)



#!/usr/bin/env python

from setuptools import setup, find_packages
from io import open

# Version info -- read without importing
_locals = {}
with open("pytest_relaxed/_version.py") as fp:
    exec(fp.read(), None, _locals)
version = _locals["__version__"]


setup(
    name="pytest-relaxed",
    version=version,
    description="Relaxed test discovery/organization for pytest",
    license="BSD",
    url="https://pytest-relaxed.readthedocs.io/",
    author="Jeff Forcier",
    author_email="jeff@bitprophet.org",
    long_description="\n" + open("README.rst", encoding="utf-8").read(),
    packages=find_packages(),
    entry_points={
        # TODO: do we need to name the LHS 'pytest_relaxed' too? meh
        "pytest11": ["relaxed = pytest_relaxed.plugin"]
    },
    install_requires=[
        # NOTE: pytest 3.3 broke something, not sure what yet
        "pytest>=3,<3.3",
        # TODO: ditto; six is so widely used it's prob worth having a broad pin
        "six>=1,<2",
        # TODO: ditto again!
        # TODO: but also, we could make this optional if it somehow ever caused
        # issues for anyone. Only necessary for @raises right now.
        "decorator>=4,<5",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python",
        "Topic :: Software Development :: Testing",
    ],
)



from pytest import skip

# Load some fixtures we expose, without actually loading our entire plugin
from pytest_relaxed.fixtures import environ  # noqa


# TODO: how best to make all of this opt-in/out? Reporter as separate plugin?
# (May not be feasible if it has to assume something about how our collection
# works?) CLI option (99% sure we can hook into that as a plugin)?


def _expect_regular_output(testdir):
    output = testdir.runpytest().stdout.str()
    results = (
        """
behaviors.py ..
other_behaviors.py s.F.
""".lstrip()
    )
    # Regular results w/ status letters
    assert results in output
    # Failure/traceback reporting
    assert "== FAILURES ==" in output
    assert "AssertionError" in output
    # Summary
    assert "== 1 failed, 4 passed, 1 skipped in " in output


class TestRegularFunctions:
    """
    Function-oriented test modules, normal display mode.
    """

    def test_acts_just_like_normal_pytest(self, testdir):
        testdir.makepyfile(
            behaviors="""
                def behavior_one():
                    pass

                def behavior_two():
                    pass
            """,
            other_behaviors="""
                from pytest import skip

                def behavior_one():
                    skip()

                def behavior_two():
                    pass

                def behavior_three():
                    assert False

                def behavior_four():
                    pass
            """,
        )
        _expect_regular_output(testdir)


class TestVerboseFunctions:
    """
    Function-oriented test modules, verbose display mode.
    """

    def test_displays_tests_indented_under_module_header(self, testdir):
        # TODO: at least, that seems like a reasonable thing to do offhand
        skip()

    def test_test_prefixes_are_stripped(self, testdir):
        testdir.makepyfile(
            legacy="""
            def test_some_things():
                pass

            def test_other_things():
                pass
        """
        )
        expected = (
            """
some things
other things
""".lstrip()
        )
        output = testdir.runpytest_subprocess("-v").stdout.str()
        assert expected in output


class TestNormalClasses:
    """
    Class-oriented test modules, normal display mode.
    """

    def acts_just_like_normal_pytest(self, testdir):
        testdir.makepyfile(
            behaviors="""
                class Behaviors:
                    def behavior_one(self):
                        pass

                    def behavior_two(self):
                        pass
            """,
            other_behaviors="""
                from pytest import skip

                class OtherBehaviors:
                    def behavior_one(self):
                        skip()

                    def behavior_two(self):
                        pass

                    def behavior_three(self):
                        assert False

                    def behavior_four(self):
                        pass
            """,
        )
        _expect_regular_output(testdir)


class TestVerboseClasses:
    """
    Class-oriented test modules, verbose display mode.
    """

    def test_shows_tests_nested_under_classes_without_files(self, testdir):
        testdir.makepyfile(
            behaviors="""
                class Behaviors:
                    def behavior_one(self):
                        pass

                    def behavior_two(self):
                        pass
            """,
            other_behaviors="""
                from pytest import skip
                class OtherBehaviors:
                    def behavior_one(self):
                        pass

                    def behavior_two(self):
                        skip()

                    def behavior_three(self):
                        pass

                    def behavior_four(self):
                        assert False
            """,
        )
        output = testdir.runpytest_subprocess("-v").stdout.str()
        results = (
            """
Behaviors

    behavior one
    behavior two

OtherBehaviors

    behavior one
    behavior two
    behavior three
    behavior four
""".lstrip()
        )
        assert results in output
        # Ensure we're not accidentally nixing failure, summary output
        assert "== FAILURES ==" in output
        assert "AssertionError" in output
        # Summary
        assert "== 1 failed, 4 passed, 1 skipped in " in output

    def test_tests_are_colorized_by_test_result(  # noqa: F811,E501
        self, testdir, environ
    ):
        # Make sure py._io.TerminalWriters write colors despite pytest output
        # capturing, which would otherwise trigger a 'False' result for "should
        # markup output".
        environ["PY_COLORS"] = "1"
        testdir.makepyfile(
            behaviors="""
                class Behaviors:
                    def behavior_one(self):
                        pass

                    def behavior_two(self):
                        pass
            """,
            other_behaviors="""
                from pytest import skip
                class OtherBehaviors:
                    def behavior_one(self):
                        pass

                    def behavior_two(self):
                        skip()

                    def behavior_three(self):
                        pass

                    def behavior_four(self):
                        assert False
            """,
        )
        output = testdir.runpytest_subprocess("-v").stdout.str()
        results = (
            """
Behaviors

    \x1b[32mbehavior one\x1b[0m
    \x1b[32mbehavior two\x1b[0m

OtherBehaviors

    \x1b[32mbehavior one\x1b[0m
    \x1b[33mbehavior two\x1b[0m
    \x1b[32mbehavior three\x1b[0m
    \x1b[31mbehavior four\x1b[0m
""".lstrip()
        )
        assert results in output
        # Ensure we're not accidentally nixing failure, summary output
        assert "== FAILURES ==" in output
        assert "AssertionError" in output
        # Summary
        assert "== 1 failed, 4 passed, 1 skipped in " in output

    def test_nests_many_levels_deep_no_problem(self, testdir):
        testdir.makepyfile(
            behaviors="""
                class Behaviors:
                    def behavior_one(self):
                        pass

                    def behavior_two(self):
                        pass

                    class MoreBehaviors:
                        def an_behavior(self):
                            pass

                        def another_behavior(self):
                            pass

                    class YetMore:
                        class Behaviors:
                            def yup(self):
                                pass

                            def still_works(self):
                                pass
            """
        )
        expected = (
            """
Behaviors

    behavior one
    behavior two

    MoreBehaviors

        an behavior
        another behavior

    YetMore

        Behaviors

            yup
            still works
""".lstrip()
        )
        assert expected in testdir.runpytest("-v").stdout.str()

    def test_headers_and_tests_have_underscores_turn_to_spaces(self, testdir):
        testdir.makepyfile(
            behaviors="""
            class some_non_class_name_like_header:
                def a_test_sentence(self):
                    pass
        """
        )
        expected = (
            """
some non class name like header

    a test sentence
""".lstrip()
        )
        assert expected in testdir.runpytest("-v").stdout.str()

    def test_test_prefixes_are_stripped(self, testdir):
        testdir.makepyfile(
            stripping="""
            class TestSomeStuff:
                def test_the_stuff(self):
                    pass
        """
        )
        expected = (
            """
SomeStuff

    the stuff

""".lstrip()
        )
        assert expected in testdir.runpytest("-v").stdout.str()

    def test_test_suffixes_are_stripped(self, testdir):
        testdir.makepyfile(
            stripping="""
            class StuffTest:
                def test_yup(self):
                    pass
        """
        )
        expected = (
            """
Stuff

    yup

""".lstrip()
        )
        assert expected in testdir.runpytest("-v").stdout.str()


class TestNormalMixed:
    """
    Mixed function and class test modules, normal display mode.
    """
    # TODO: currently undefined; spec never even really worked for this
    pass


class TestVerboseMixed:
    """
    Mixed function and class test modules, verbose display mode.
    """
    # TODO: currently undefined; spec never even really worked for this
    pass
